@@ -5,10 +5,10 @@
 package lsp
 
 import (
-	"bytes"
 	"context"
 	"encoding/json"
 	"fmt"
+	"go/build"
 	"log"
 	"os"
 	"path"
@@ -17,16 +17,19 @@ import (
 	"strings"
 	"sync"
 
+	"golang.org/x/tools/gopls/internal/bug"
 	"golang.org/x/tools/gopls/internal/lsp/debug"
 	"golang.org/x/tools/gopls/internal/lsp/protocol"
 	"golang.org/x/tools/gopls/internal/lsp/source"
 	"golang.org/x/tools/gopls/internal/span"
-	"golang.org/x/tools/internal/bug"
 	"golang.org/x/tools/internal/event"
 	"golang.org/x/tools/internal/jsonrpc2"
 )
 
 func (s *Server) initialize(ctx context.Context, params *protocol.ParamInitialize) (*protocol.InitializeResult, error) {
+	ctx, done := event.Start(ctx, "lsp.Server.initialize")
+	defer done()
+
 	s.stateMu.Lock()
 	if s.state >= serverInitializing {
 		defer s.stateMu.Unlock()
@@ -57,21 +60,21 @@ func (s *Server) initialize(ctx context.Context, params *protocol.ParamInitializ
 	if err := s.handleOptionResults(ctx, source.SetOptions(options, params.InitializationOptions)); err != nil {
 		return nil, err
 	}
-	options.ForClientCapabilities(params.Capabilities)
+	options.ForClientCapabilities(params.ClientInfo, params.Capabilities)
 
 	if options.ShowBugReports {
 		// Report the next bug that occurs on the server.
-		bugCh := bug.Notify()
-		go func() {
-			b := <-bugCh
+		bug.Handle(func(b bug.Bug) {
 			msg := &protocol.ShowMessageParams{
 				Type:    protocol.Error,
 				Message: fmt.Sprintf("A bug occurred on the server: %s\nLocation:%s", b.Description, b.Key),
 			}
-			if err := s.eventuallyShowMessage(context.Background(), msg); err != nil {
-				log.Printf("error showing bug: %v", err)
-			}
-		}()
+			go func() {
+				if err := s.eventuallyShowMessage(context.Background(), msg); err != nil {
+					log.Printf("error showing bug: %v", err)
+				}
+			}()
+		})
 	}
 
 	folders := params.WorkspaceFolders
@@ -107,7 +110,7 @@ func (s *Server) initialize(ctx context.Context, params *protocol.ParamInitializ
 		}
 	}
 	var renameOpts interface{} = true
-	if r := params.Capabilities.TextDocument.Rename; r.PrepareSupport {
+	if r := params.Capabilities.TextDocument.Rename; r != nil && r.PrepareSupport {
 		renameOpts = protocol.RenameOptions{
 			PrepareProvider: r.PrepareSupport,
 		}
@@ -137,46 +140,55 @@ See https://github.com/golang/go/issues/45732 for more information.`,
 
 	return &protocol.InitializeResult{
 		Capabilities: protocol.ServerCapabilities{
-			CallHierarchyProvider: true,
+			CallHierarchyProvider: &protocol.Or_ServerCapabilities_callHierarchyProvider{Value: true},
 			CodeActionProvider:    codeActionProvider,
 			CodeLensProvider:      &protocol.CodeLensOptions{}, // must be non-nil to enable the code lens capability
-			CompletionProvider: protocol.CompletionOptions{
+			CompletionProvider: &protocol.CompletionOptions{
 				TriggerCharacters: []string{"."},
 			},
-			DefinitionProvider:         true,
-			TypeDefinitionProvider:     true,
-			ImplementationProvider:     true,
-			DocumentFormattingProvider: true,
-			DocumentSymbolProvider:     true,
-			WorkspaceSymbolProvider:    true,
-			ExecuteCommandProvider: protocol.ExecuteCommandOptions{
-				Commands: options.SupportedCommands,
+			DefinitionProvider:         &protocol.Or_ServerCapabilities_definitionProvider{Value: true},
+			TypeDefinitionProvider:     &protocol.Or_ServerCapabilities_typeDefinitionProvider{Value: true},
+			ImplementationProvider:     &protocol.Or_ServerCapabilities_implementationProvider{Value: true},
+			DocumentFormattingProvider: &protocol.Or_ServerCapabilities_documentFormattingProvider{Value: true},
+			DocumentSymbolProvider:     &protocol.Or_ServerCapabilities_documentSymbolProvider{Value: true},
+			WorkspaceSymbolProvider:    &protocol.Or_ServerCapabilities_workspaceSymbolProvider{Value: true},
+			ExecuteCommandProvider: &protocol.ExecuteCommandOptions{
+				Commands: nonNilSliceString(options.SupportedCommands),
 			},
-			FoldingRangeProvider:      true,
-			HoverProvider:             true,
-			DocumentHighlightProvider: true,
-			DocumentLinkProvider:      protocol.DocumentLinkOptions{},
+			FoldingRangeProvider:      &protocol.Or_ServerCapabilities_foldingRangeProvider{Value: true},
+			HoverProvider:             &protocol.Or_ServerCapabilities_hoverProvider{Value: true},
+			DocumentHighlightProvider: &protocol.Or_ServerCapabilities_documentHighlightProvider{Value: true},
+			DocumentLinkProvider:      &protocol.DocumentLinkOptions{},
 			InlayHintProvider:         protocol.InlayHintOptions{},
-			ReferencesProvider:        true,
+			ReferencesProvider:        &protocol.Or_ServerCapabilities_referencesProvider{Value: true},
 			RenameProvider:            renameOpts,
-			SignatureHelpProvider: protocol.SignatureHelpOptions{
+			SelectionRangeProvider:    &protocol.Or_ServerCapabilities_selectionRangeProvider{Value: true},
+			SemanticTokensProvider: protocol.SemanticTokensOptions{
+				Range: &protocol.Or_SemanticTokensOptions_range{Value: true},
+				Full:  &protocol.Or_SemanticTokensOptions_full{Value: true},
+				Legend: protocol.SemanticTokensLegend{
+					TokenTypes:     nonNilSliceString(s.session.Options().SemanticTypes),
+					TokenModifiers: nonNilSliceString(s.session.Options().SemanticMods),
+				},
+			},
+			SignatureHelpProvider: &protocol.SignatureHelpOptions{
 				TriggerCharacters: []string{"(", ","},
 			},
 			TextDocumentSync: &protocol.TextDocumentSyncOptions{
 				Change:    protocol.Incremental,
 				OpenClose: true,
-				Save: protocol.SaveOptions{
+				Save: &protocol.SaveOptions{
 					IncludeText: false,
 				},
 			},
-			Workspace: protocol.Workspace6Gn{
-				WorkspaceFolders: protocol.WorkspaceFolders5Gn{
+			Workspace: &protocol.Workspace6Gn{
+				WorkspaceFolders: &protocol.WorkspaceFolders5Gn{
 					Supported:           true,
 					ChangeNotifications: "workspace/didChangeWorkspaceFolders",
 				},
 			},
 		},
-		ServerInfo: protocol.PServerInfoMsg_initialize{
+		ServerInfo: &protocol.PServerInfoMsg_initialize{
 			Name:    "gopls",
 			Version: string(goplsVersion),
 		},
@@ -184,6 +196,9 @@ See https://github.com/golang/go/issues/45732 for more information.`,
 }
 
 func (s *Server) initialized(ctx context.Context, params *protocol.InitializedParams) error {
+	ctx, done := event.Start(ctx, "lsp.Server.initialized")
+	defer done()
+
 	s.stateMu.Lock()
 	if s.state >= serverInitialized {
 		defer s.stateMu.Unlock()
@@ -213,9 +228,6 @@ func (s *Server) initialized(ctx context.Context, params *protocol.InitializedPa
 			Method: "workspace/didChangeConfiguration",
 		})
 	}
-	if options.SemanticTokens && options.DynamicRegistrationSemanticTokensSupported {
-		registrations = append(registrations, semanticTokenRegistration(options.SemanticTypes, options.SemanticMods))
-	}
 	if len(registrations) > 0 {
 		if err := s.client.RegisterCapability(ctx, &protocol.RegistrationParams{
 			Registrations: registrations,
@@ -228,14 +240,16 @@ func (s *Server) initialized(ctx context.Context, params *protocol.InitializedPa
 
 // GoVersionTable maps Go versions to the gopls version in which support will
 // be deprecated, and the final gopls version supporting them without warnings.
-// Keep this in sync with gopls/README.md
+// Keep this in sync with gopls/README.md.
 //
 // Must be sorted in ascending order of Go version.
 //
 // Mutable for testing.
 var GoVersionTable = []GoVersionSupport{
 	{12, "", "v0.7.5"},
-	{15, "v0.11.0", "v0.9.5"},
+	{15, "", "v0.9.5"},
+	{16, "v0.13.0", "v0.11.0"},
+	{17, "v0.13.0", "v0.11.0"},
 }
 
 // GoVersionSupport holds information about end-of-life Go version support.
@@ -251,11 +265,13 @@ func OldestSupportedGoVersion() int {
 	return GoVersionTable[len(GoVersionTable)-1].GoVersion + 1
 }
 
-// versionMessage returns the warning/error message to display if the user is
-// on the given Go version, if any. The goVersion variable is the X in Go 1.X.
+// versionMessage returns the warning/error message to display if the user has
+// the given Go version, if any. The goVersion variable is the X in Go 1.X. If
+// fromBuild is set, the Go version is the version used to build gopls.
+// Otherwise, it is the go command version.
 //
 // If goVersion is invalid (< 0), it returns "", 0.
-func versionMessage(goVersion int) (string, protocol.MessageType) {
+func versionMessage(goVersion int, fromBuild bool) (string, protocol.MessageType) {
 	if goVersion < 0 {
 		return "", 0
 	}
@@ -265,7 +281,11 @@ func versionMessage(goVersion int) (string, protocol.MessageType) {
 			var msgBuilder strings.Builder
 
 			mType := protocol.Error
-			fmt.Fprintf(&msgBuilder, "Found Go version 1.%d", goVersion)
+			if fromBuild {
+				fmt.Fprintf(&msgBuilder, "Gopls was built with Go version 1.%d", goVersion)
+			} else {
+				fmt.Fprintf(&msgBuilder, "Found Go version 1.%d", goVersion)
+			}
 			if v.DeprecatedVersion != "" {
 				// not deprecated yet, just a warning
 				fmt.Fprintf(&msgBuilder, ", which will be unsupported by gopls %s. ", v.DeprecatedVersion)
@@ -288,15 +308,15 @@ func versionMessage(goVersion int) (string, protocol.MessageType) {
 //
 // It should be called after views change.
 func (s *Server) checkViewGoVersions() {
-	oldestVersion := -1
+	oldestVersion, fromBuild := go1Point(), true
 	for _, view := range s.session.Views() {
 		viewVersion := view.GoVersion()
 		if oldestVersion == -1 || viewVersion < oldestVersion {
-			oldestVersion = viewVersion
+			oldestVersion, fromBuild = viewVersion, false
 		}
 	}
 
-	if msg, mType := versionMessage(oldestVersion); msg != "" {
+	if msg, mType := versionMessage(oldestVersion, fromBuild); msg != "" {
 		s.eventuallyShowMessage(context.Background(), &protocol.ShowMessageParams{
 			Type:    mType,
 			Message: msg,
@@ -304,22 +324,37 @@ func (s *Server) checkViewGoVersions() {
 	}
 }
 
+// go1Point returns the x in Go 1.x. If an error occurs extracting the go
+// version, it returns -1.
+//
+// Copied from the testenv package.
+func go1Point() int {
+	for i := len(build.Default.ReleaseTags) - 1; i >= 0; i-- {
+		var version int
+		if _, err := fmt.Sscanf(build.Default.ReleaseTags[i], "go1.%d", &version); err != nil {
+			continue
+		}
+		return version
+	}
+	return -1
+}
+
 func (s *Server) addFolders(ctx context.Context, folders []protocol.WorkspaceFolder) error {
 	originalViews := len(s.session.Views())
 	viewErrors := make(map[span.URI]error)
 
-	var wg sync.WaitGroup
+	var ndiagnose sync.WaitGroup // number of unfinished diagnose calls
 	if s.session.Options().VerboseWorkDoneProgress {
 		work := s.progress.Start(ctx, DiagnosticWorkTitle(FromInitialWorkspaceLoad), "Calculating diagnostics for initial workspace load...", nil, nil)
 		defer func() {
 			go func() {
-				wg.Wait()
+				ndiagnose.Wait()
 				work.End(ctx, "Done.")
 			}()
 		}()
 	}
 	// Only one view gets to have a workspace.
-	var allFoldersWg sync.WaitGroup
+	var nsnapshots sync.WaitGroup // number of unfinished snapshot initializations
 	for _, folder := range folders {
 		uri := span.URIFromURI(folder.URI)
 		// Ignore non-file URIs.
@@ -338,41 +373,31 @@ func (s *Server) addFolders(ctx context.Context, folders []protocol.WorkspaceFol
 		}
 		// Inv: release() must be called once.
 
-		var swg sync.WaitGroup
-		swg.Add(1)
-		allFoldersWg.Add(1)
-		// TODO(adonovan): this looks fishy. Is AwaitInitialized
-		// supposed to be called once per folder?
+		// Initialize snapshot asynchronously.
+		initialized := make(chan struct{})
+		nsnapshots.Add(1)
 		go func() {
-			defer swg.Done()
-			defer allFoldersWg.Done()
 			snapshot.AwaitInitialized(ctx)
 			work.End(ctx, "Finished loading packages.")
+			nsnapshots.Done()
+			close(initialized) // signal
 		}()
 
-		// Print each view's environment.
-		buf := &bytes.Buffer{}
-		if err := snapshot.WriteEnv(ctx, buf); err != nil {
-			viewErrors[uri] = err
-			release()
-			continue
-		}
-		event.Log(ctx, buf.String())
-
-		// Diagnose the newly created view.
-		wg.Add(1)
+		// Diagnose the newly created view asynchronously.
+		ndiagnose.Add(1)
 		go func() {
-			s.diagnoseDetached(snapshot)
-			swg.Wait()
+			s.diagnoseSnapshot(snapshot, nil, false, 0)
+			<-initialized
 			release()
-			wg.Done()
+			ndiagnose.Done()
 		}()
 	}
 
+	// Wait for snapshots to be initialized so that all files are known.
+	// (We don't need to wait for diagnosis to finish.)
+	nsnapshots.Wait()
+
 	// Register for file watching notifications, if they are supported.
-	// Wait for all snapshots to be initialized first, since all files might
-	// not yet be known to the snapshots.
-	allFoldersWg.Wait()
 	if err := s.updateWatchedDirectories(ctx); err != nil {
 		event.Error(ctx, "failed to register for file watching notifications", err)
 	}
@@ -444,18 +469,18 @@ func equalURISet(m1, m2 map[string]struct{}) bool {
 
 // registerWatchedDirectoriesLocked sends the workspace/didChangeWatchedFiles
 // registrations to the client and updates s.watchedDirectories.
+// The caller must not subsequently mutate patterns.
 func (s *Server) registerWatchedDirectoriesLocked(ctx context.Context, patterns map[string]struct{}) error {
 	if !s.session.Options().DynamicWatchedFilesSupported {
 		return nil
 	}
-	for k := range s.watchedGlobPatterns {
-		delete(s.watchedGlobPatterns, k)
-	}
-	var watchers []protocol.FileSystemWatcher
+	s.watchedGlobPatterns = patterns
+	watchers := make([]protocol.FileSystemWatcher, 0, len(patterns)) // must be a slice
+	val := protocol.WatchChange | protocol.WatchDelete | protocol.WatchCreate
 	for pattern := range patterns {
 		watchers = append(watchers, protocol.FileSystemWatcher{
 			GlobPattern: pattern,
-			Kind:        uint32(protocol.WatchChange + protocol.WatchDelete + protocol.WatchCreate),
+			Kind:        &val,
 		})
 	}
 
@@ -471,10 +496,6 @@ func (s *Server) registerWatchedDirectoriesLocked(ctx context.Context, patterns
 		return err
 	}
 	s.watchRegistrationCount++
-
-	for k, v := range patterns {
-		s.watchedGlobPatterns[k] = v
-	}
 	return nil
 }
 
@@ -559,7 +580,7 @@ func (s *Server) handleOptionResults(ctx context.Context, results source.OptionR
 // We don't want to return errors for benign conditions like wrong file type,
 // so callers should do if !ok { return err } rather than if err != nil.
 // The returned cleanup function is non-nil even in case of false/error result.
-func (s *Server) beginFileRequest(ctx context.Context, pURI protocol.DocumentURI, expectKind source.FileKind) (source.Snapshot, source.VersionedFileHandle, bool, func(), error) {
+func (s *Server) beginFileRequest(ctx context.Context, pURI protocol.DocumentURI, expectKind source.FileKind) (source.Snapshot, source.FileHandle, bool, func(), error) {
 	uri := pURI.SpanURI()
 	if !uri.IsFile() {
 		// Not a file URI. Stop processing the request, but don't return an error.
@@ -569,8 +590,11 @@ func (s *Server) beginFileRequest(ctx context.Context, pURI protocol.DocumentURI
 	if err != nil {
 		return nil, nil, false, func() {}, err
 	}
-	snapshot, release := view.Snapshot(ctx)
-	fh, err := snapshot.GetVersionedFile(ctx, uri)
+	snapshot, release, err := view.Snapshot()
+	if err != nil {
+		return nil, nil, false, func() {}, err
+	}
+	fh, err := snapshot.ReadFile(ctx, uri)
 	if err != nil {
 		release()
 		return nil, nil, false, func() {}, err
@@ -586,6 +610,9 @@ func (s *Server) beginFileRequest(ctx context.Context, pURI protocol.DocumentURI
 // shutdown implements the 'shutdown' LSP handler. It releases resources
 // associated with the server and waits for all ongoing work to complete.
 func (s *Server) shutdown(ctx context.Context) error {
+	ctx, done := event.Start(ctx, "lsp.Server.shutdown")
+	defer done()
+
 	s.stateMu.Lock()
 	defer s.stateMu.Unlock()
 	if s.state < serverInitialized {
@@ -605,6 +632,9 @@ func (s *Server) shutdown(ctx context.Context) error {
 }
 
 func (s *Server) exit(ctx context.Context) error {
+	ctx, done := event.Start(ctx, "lsp.Server.exit")
+	defer done()
+
 	s.stateMu.Lock()
 	defer s.stateMu.Unlock()
 
@@ -614,7 +644,35 @@ func (s *Server) exit(ctx context.Context) error {
 		// TODO: We should be able to do better than this.
 		os.Exit(1)
 	}
-	// we don't terminate the process on a normal exit, we just allow it to
+	// We don't terminate the process on a normal exit, we just allow it to
 	// close naturally if needed after the connection is closed.
 	return nil
 }
+
+// TODO: when we can assume go1.18, replace with generic
+// (after retiring support for go1.17)
+func nonNilSliceString(x []string) []string {
+	if x == nil {
+		return []string{}
+	}
+	return x
+}
+func nonNilSliceTextEdit(x []protocol.TextEdit) []protocol.TextEdit {
+	if x == nil {
+		return []protocol.TextEdit{}
+	}
+
+	return x
+}
+func nonNilSliceCompletionItemTag(x []protocol.CompletionItemTag) []protocol.CompletionItemTag {
+	if x == nil {
+		return []protocol.CompletionItemTag{}
+	}
+	return x
+}
+func emptySliceDiagnosticTag(x []protocol.DiagnosticTag) []protocol.DiagnosticTag {
+	if x == nil {
+		return []protocol.DiagnosticTag{}
+	}
+	return x
+}
